#include "neighbor.h"
#include "libheader.h"
#include "header.h"
#include "table.h"

extern unsigned long oneMask[];
extern unsigned long twoMask[];
extern unsigned long threeMask[];

/*****************************************************************************/
/*                                                                           */
/* Constructor and Destructor                                                */
/*                                                                           */
/*****************************************************************************/
Neighbor::Neighbor(const long motifLength, const long distance) {
  this->motifLength=motifLength;
  this->maxHamDist=distance;
  this->maskList=NULL;
  this->maskNoArr=NULL;
}

Neighbor::~Neighbor() {
  if(maskList!=NULL){
    for(long i=0;i<=maxHamDist;i++) {
      if(maskList[i]!=NULL)
        delete[] maskList[i];
    }
    delete[] maskList;
  }
  if(maskNoArr!=NULL)
    delete[] maskNoArr;
}

/*****************************************************************************/
/*                                                                           */
/* Pub 1 Access Attributes                                                   */
/*                                                                           */
/*****************************************************************************/
const unsigned long* Neighbor::getMaskArr(long d) {
  if(d<=maxHamDist && d>=0)
    return maskList[d];
  else
    return NULL;
}

long Neighbor::getMaskNo(long d) {
  if(d<=maxHamDist && d>=0)
    return maskNoArr[d];
  else
    return 0;
}
/*****************************************************************************/
/*                                                                           */
/* Pub 2 Algorithms                                                          */
/*                                                                           */
/*****************************************************************************/

void Neighbor::createNeighborList() {
  if(maskList!=NULL){
    for(long i=0;i<=maxHamDist;i++) {
      if(maskList[i]!=NULL)
        delete[] maskList[i];
    }
    delete[] maskList;
  }
  if(maskNoArr!=NULL)
    delete[] maskNoArr;
  
  maskList=new unsigned long*[maxHamDist+1];
  maskNoArr=new long[maxHamDist+1];
  
  //hamming distance=0;
  maskList[0]=new unsigned long[1];
  maskList[0][0]=0x0;
  maskNoArr[0]=1;
  
  //for each hamming distance;
  for(long d=1;d<=min(motifLength, maxHamDist);d++) {
    
    long* combinationArr=new long[d];
    vector<long*> comArrVec;
    Tools::combination(combinationArr, d, 0, motifLength, comArrVec);
    
    long arrSize=(long)(comArrVec.size()*pow(3.0,(int)d));
    maskList[d]=new unsigned long[arrSize];
    maskNoArr[d]=arrSize;
    
    //try if we can create the array
    if(maskList[d]==NULL) {
      cout<<"In Neighbor::createNeighborList"<<endl;
      cout<<"Not enough memory for create maskList["<<d<<"]"<<endl;
      exit(1);
    }
    
    long maskNo=0;
    for(long i=0;i<(long)comArrVec.size();i++)
      createMask(comArrVec.at(i), d, maskList[d], 0x0, 0, maskNo);
    
    //disallocate memory
    delete[] combinationArr;
    for(long d=0;d<(long)comArrVec.size();d++)
      delete[] comArrVec.at(d);
  }
  
  for(long d=min(motifLength, maxHamDist)+1;d<=maxHamDist;d++){
    maskList[d]=0;
    maskNoArr[d]=0;
  }
  
}

/*****************************************************************************/
/*                                                                           */
/* Pri 1 Algorithms                                                          */
/*                                                                           */
/*****************************************************************************/
void Neighbor::createMask(const long* combinationArr, const long comArrSize
                          , unsigned long* maskArr, const unsigned long mask
                          , const long pos, long &maskNo) {
  if(pos==comArrSize-1) {
    maskArr[maskNo]= mask | oneMask[combinationArr[pos]];
    maskNo++;
    maskArr[maskNo]= mask | twoMask[combinationArr[pos]];
    maskNo++;
    maskArr[maskNo]= mask | threeMask[combinationArr[pos]];
    maskNo++;
  } else {
    createMask(combinationArr, comArrSize, maskArr
               , mask | oneMask[combinationArr[pos]]
               , pos+1, maskNo);
    createMask(combinationArr, comArrSize, maskArr
               , mask | twoMask[combinationArr[pos]]
               , pos+1, maskNo);
    createMask(combinationArr, comArrSize, maskArr
               , mask | threeMask[combinationArr[pos]]
               , pos+1, maskNo);
  }
}

/*
int main(){
  int w=8;
  int d=4;
  
  Neighbor neighbor(w, d);
  neighbor.createNeighborList();
  const unsigned long* maskArr=neighbor.getMaskArr(d);
  
  char* pattern=new char[w+1];
  long unsigned originalPattern=0;
  
  Tools::longToPattern(pattern,w,originalPattern);
  cout<<"The original pattern is"<<endl;
  cout<<pattern<<endl;
  cout<<endl;
  
  long arrSize=neighbor.getMaskNo(d);
  
  for(long i=0;i<arrSize;i++) {
    Tools::longToPattern(pattern,w, maskArr[i] ^ originalPattern);
    cout<<i<<"\t"<<pattern<<endl;
  }
  
  delete[] pattern;
}
*/
