#include "header.h"

/*****************************************************************************/
/*                                                                           */
/* Constructor and Destructor                                                */
/*                                                                           */
/*****************************************************************************/
DNA::DNA(const char* name, const char* sequence) {
  this->dnaName=name;
  this->dnaSeq=sequence;
  
  int dnaLength=strlen(this->dnaSeq);
  char* tmpSeqR=new char[dnaLength+1];
  
  for(int i=0;i<dnaLength;i++) {
    switch(this->dnaSeq[i]) {
      case 'A':
        tmpSeqR[dnaLength-i-1]='T';
        break;
      case 'C':
        tmpSeqR[dnaLength-i-1]='G';
        break;
      case 'G':
        tmpSeqR[dnaLength-i-1]='C';
        break;
      case 'T':
        tmpSeqR[dnaLength-i-1]='A';
        break;
      
      default:
        cout<<"In DNA::DNA"<<endl;
        cout<<"The "<<i<<"-th character in "<<name<<" is \'"<<this->dnaSeq[i]
            <<"\' which is not a valid character"<<endl;
        exit(1);
        break;
    }
  }
  tmpSeqR[dnaLength]=0;
  this->dnaSeqR=tmpSeqR;
  
  scoreArr=new double[dnaLength];
  scoreArrR=new double[dnaLength];
}

DNA::~DNA() {
  if(dnaName!=NULL)
    delete [] dnaName;
  if(dnaSeq!=NULL)
    delete [] dnaSeq;
  if(dnaSeqR!=NULL)
    delete [] dnaSeqR;
  if(scoreArr!=NULL)
    delete [] scoreArr;
  if(scoreArrR!=NULL)
    delete [] scoreArrR;
}

/*****************************************************************************/
/*                                                                           */
/* Pub 1 Access Attributes                                                   */
/*                                                                           */
/*****************************************************************************/

//get the name of the dna
const char* DNA::getName() {
  return dnaName;
}
  
//get the DNA Sequence
const char* DNA::getSeq() {
  return dnaSeq;
}

//get the reversed DNA Sequence
const char* DNA::getSeqR() {
  return dnaSeqR;
}

//get the Score array for all substrings in the DNA sequence
const double* DNA::getScoreArr() {
  return scoreArr;
}

//get the Score array for all substrings in the reversed DNA sequence
const double* DNA::getScoreArrR() {
  return scoreArrR;
}

/*****************************************************************************/
/*                                                                           */
/* Pub 2 Algorithms                                                          */
/*                                                                           */
/*****************************************************************************/

//Calculate log likelihood of all substrings
void DNA::calculateScore(const double* probMat, const int motifLength, const bool testRer) {    
  int noOfSubstring=strlen(dnaSeq)-motifLength+1;
  
  //original sequence
  for(int i=0;i<noOfSubstring;i++) {
    scoreArr[i]=calLogProbMat(probMat,&dnaSeq[i],motifLength);
  }
  //reversed sequence
  if(testRer) {
    for(int i=0;i<noOfSubstring;i++) {
      scoreArrR[i]=calLogProbMat(probMat,&dnaSeqR[i],motifLength);
    }
  }
}

//Find if a pattern exist in the dna sequence with Hamming distance <= distance
int DNA::findSubstring(const char* pattern, const int distance, const int stringLength
                       , const bool testRev) {  
  int noOfSubstring=strlen(dnaSeq)-stringLength+1;
  int existTime=0;
  int count;
  
  for(int pos=0;pos<noOfSubstring;pos++) {
    count=0;
    for(int i=0;i<stringLength;i++) {
      if(dnaSeq[pos+i]!=pattern[i]) {
        count++;
        if(count>distance)
          break;
      }
    }
    if(count<=distance) {
      existTime++;
    }
  }
  
  if(testRev) {
    for(int pos=0;pos<noOfSubstring;pos++) {
      count=0;
      for(int i=0;i<stringLength;i++) {
        if(dnaSeqR[pos+i]!=pattern[pos]) {
          count++;
          if(count>distance)
            break;
        }
      }
      if(count<=distance) {
        existTime++;
      }
    }
  }
  
  return existTime;
}
/*****************************************************************************/
/*                                                                           */
/* Pri 1 Algorithms                                                          */
/*                                                                           */
/*****************************************************************************/

//Calculation log probability of a substring generated from a matrix probMat
double DNA::calLogProbMat(const double* probMat, const char* bindingSite, const int motifLength) {
  double prob=0.0;
  for(int i=0;i<motifLength;i++) {
    switch(bindingSite[i]) {
      case 'A':
        prob+=log(probMat[i]);
        break;
      case 'C':
        prob+=log(probMat[i+motifLength]);
        break;
      case 'G':
        prob+=log(probMat[i+motifLength*2]);
        break;
      case 'T':
        prob+=log(probMat[i+motifLength*3]);
        break;
      default:
        cout<<"In DNA::calProbMat"<<endl;
        cout<<"The "<<i<<"-th character in the substring "<<bindingSite<<" is \'"<<bindingSite[i]
            <<"\' which is not a valid character"<<endl;
        exit(1);
        break;
    }
  }
  return prob;
}

//Calculation log probability of a substring generated from a background probability
double DNA::calLogBackProb(const double* backgroundMat, const char* bindingSite, const int motifLength) {
  double prob=0.0;
  for(int i=0;i<motifLength;i++) {
    switch(bindingSite[i]) {
      case 'A':
        prob+=log(backgroundMat[0]);
        break;
      case 'C':
        prob+=log(backgroundMat[1]);
        break;
      case 'G':
        prob+=log(backgroundMat[2]);
        break;
      case 'T':
        prob+=log(backgroundMat[3]);
        break;
      default:
        cout<<"In DNA::calBackProb"<<endl;
        cout<<"The "<<i<<"-th character in the substring "<<bindingSite<<" is \'"<<bindingSite[i]
            <<"\' which is not a valid character"<<endl;
        exit(1);
        break;
    }
  }
  return prob;
}


void DNA::findBindSite(const char* pattern, const long hamDist
                       , vector<BindSite*> &bindSiteVec, const bool testRev) {
  long motifLength=strlen(pattern);
  for(int i=0;i<=(int)strlen(dnaSeq)-motifLength;i++) {
    int error=0;
    for(int j=0;j<motifLength;j++) {
      if(pattern[j]!=dnaSeq[i+j])
        error++;
    }
    if(error <=hamDist)
      bindSiteVec.push_back(new BindSite(i,true));
  }
   
  if(testRev) {
    for(int i=0;i<=(int)strlen(dnaSeqR)-motifLength;i++) {
      int error=0;
      for(int j=0;j<motifLength;j++) {
        if(pattern[j]!=dnaSeqR[i+j])
          error++;
      }
      if(error <=hamDist)
        bindSiteVec.push_back(new BindSite(i,false));
    }
  }
}

short DNA::findBindSiteNum(const char* pattern, const long hamDist
                           , const bool testRev) {
  long motifLength=strlen(pattern);
  short count=0;
  for(int i=0;i<=(int)strlen(dnaSeq)-motifLength;i++) {
    int error=0;
    for(int j=0;j<motifLength;j++) {
      if(pattern[j]!=dnaSeq[i+j])
        error++;
    }
    if(error <=hamDist)
      count++;
  }
  if(testRev) {
    for(int i=0;i<=(int)strlen(dnaSeqR)-motifLength;i++) {
      int error=0;
      for(int j=0;j<motifLength;j++) {
        if(pattern[j]!=dnaSeqR[i+j])
          error++;
      }
      if(error <=hamDist)
        count++;
    }
  }
  return count;
}
